from .MedNeXt import MedNeXt
from .MultiResUNet import MultiResUNet
from .TinyUNet import TinyUNet
from .UNet import UNet
from .UNetR import UNetR
from .UXNet import UXNet

networks = [MedNeXt, MultiResUNet, TinyUNet, UNet, UNetR, UXNet]
